/*!
* \file Embankment.h
* \brief Calculation embankment attenuation
* \author CSTB
* \version 1.0
*/

#ifndef EMBANKMENT_H
#define EMBANKMENT_H

#include "../PathStructures.h"

#ifndef _WIN32
#define  Embankment   _Local_PROPAN8_Embankment_
#endif

namespace EmbankmentNMPB
{
	/*!
	 * \brief Class used to calculate the embankment attenuation
	 *
	 * p. 43-44 -  9.3.5
	 *
	 */
	class Embankment
	{
	public:
		
		/*!
		* \brief Initialization of the Embankment class and embankment attenuation calculation
		* 
		* \param terrainItems the terrain elements
		* 
		* \param freq the frequency
		* 
		* \throws ERRFrequency
		*/
		Embankment(vector<ProfilePointNMPB*> terrainItems, double freq);

		/*!
		 * \brief Embankment Destructor
		 */
		~Embankment(void)
		{
			if(_source) delete _source;
			if(_receiver) delete _receiver;
			if(_platformEnd) delete _platformEnd;
			if(_O1) delete _O1;
			if(_O2) delete _O2;
			if(_sourceImage) delete _sourceImage;
		}

		/*!
		 * \brief Gets the calculated embankment attenuation
		 * 
		 * \returns the embankment attenuation
		 */
		double GetEmbankmentAttenuation()
		{
			return _embankmentAttenuation;
		}

		/*!
		 * \brief Gets the calculated source image 2D Position
		 * 
		 * \returns the source image 2D Position
		 */
		Position2D* GetSourceImage()
		{
			return _sourceImage;
		}

	protected:
		/*!
		* \brief the source position (S)
		*/
		Position2D* _source;
		/*!
		* \brief the receiver position (R)
		*/
		Position2D* _receiver;
		/*!
		* \brief the platform end
		*/
		Position2D* _platformEnd;
		/*!
		* \brief the foot of the bank position
		*/
		Position2D* _O1;
		/*!
		* \brief the top of the bank position
		*/
		Position2D* _O2;
		/*!
		* \brief the source image position (S'')
		*/
		Position2D* _sourceImage;
		/*!
		* \brief the (O1x,O1O2) angle
		*/
		double _angleAlpha;
		/*!
		* \brief the slope impedance
		*/
		double _Gembankment;
		/*!
		 * \brief the embankment attenuation
		 */
		double _embankmentAttenuation;

		/*!
		* \brief Calculates the 2D position in the perpendicular plane to road
		* 
		* \param posToFill The position the fill
		* 
		* \param posToRead The position to read
		* 
		* \param cosTheta cos(Theta) 
		*/
		void FillPosInPerpendicularPlane(Position2D* posToFill, Position2D const* posToRead, double cosTheta);

		/*!
		* \brief Calculates the 2D position of needed items in the perpendicular plane to road
		*	and check data
		* 
		* \param terrainItems The terrain items
		* 
		* \returns True if the data are OK
		* 
		* \throws ERREmbankment
		*/
		bool DataInPerpendicularPlane(vector<ProfilePointNMPB*> terrainItems);

		/*!
		* \brief Checks the data
		*
		*	Checking according  9.3.5 p. 43
		*
		* \param terrainItems The terrain items
		* 
		* \returns true if embankment can be calculated
		*/
		bool CheckData(vector<ProfilePointNMPB*> terrainItems);

		/*!
		* \brief Calculates line coefficients : z = a*d + b (or d = b if line vertical)
		*	Warning The 2 positions must be different
		*
		* \param pos1 The first line point 
		* 
		* \param pos2 The second line pont
		* 
		* \param aCoef The a coefficient to fill
		* 
		* \param bCoef The b coefficient to fill
		* 
		* \returns true if the line is vertical
		*/
		bool LineCoeff(Position2D const* pos1, Position2D const* pos2, double &aCoef, double &bCoef);

		/*!
		* \brief Calculates the source image by the line z = a * d + b (or d = b if vertical)
		* 
		* \param aCoef The a coefficient
		* 
		* \param bCoef The b coefficient
		* 
		* \param verticalSlope true if vertical line
		*/
		void SourceImage(double aCoef, double bCoef, bool verticalSlope);

		/*!
		* \brief Calculates the lines intersection
		* 
		* \param a1 Parameter a1 of the first line
		* 
		* \param b1 Parameter b1 of the first line
		* 
		* \param vert1 True if the first line is vertical
		* 
		* \param a2 Parameter a2 of the second line
		* 
		* \param b2 Parameter b2 of the second line
		* 
		* \param vert2 True if the second line is vertical
		* 
		* \param Ipoint The intersection point to fill
		* 
		* \returns true if there is an only intersection
		*/
		bool LinesIntersection(double a1, double b1, bool vert1, double a2, double b2, bool vert2, Position2D* Ipoint);

		/*!
		* \brief Calculates the Theta angle in the formula (28) p.43
		* 
		* \returns the angle
		*/
		double ThetaAngle();

		/*!
		* \brief Calculates half-length e
		* 
		* p. 43 -  9.3.5 formula (28)
		*
		* \param cosAngleTheta The theta cosines (must be != 0)
		* 
		* \param freq The frequency (must be > 0)
		* 
		* \param SimageI The distance between Simage and I
		* 
		* \param IR The distance between Receiver and I
		* 
		* \returns the half-length e
		*/
		double HalfLength(double cosAngleTheta, double freq, double SimageI, double IR);

		/*!
		* \brief Calculates embankment attenuation (in _embankmentAttenuation)
		* 
		* \param intersect The intersection point (I)
		* 
		* \param freq The frequency
		*/ 
		void EmbankmentCalculation(Position2D const* intersect, double freq);
	};
}

#endif