#ifndef _RAILWAYDATABASE_INCLUDED_
#define _RAILWAYDATABASE_INCLUDED_

#include "RailwayEmissionNMPB08.h"
#include "../LectureXML/LectureXML.hpp"
#include <string>
#include <vector>
#include <map>
#include <assert.h>

// ----------------------------------------------------------------------------------
/*!
 *	\typedef IDREF
 *  \brief storage type for identifier/references in XML files
 */
typedef std::string IDREF ;

// ----------------------------------------------------------------------------------
/*!
 *	\brief storage for spectral data
 *	       this simplified version assuming all data is stored as 18 third octave bands 
 *	       in the range 100 - 5000 Hz
 */

// on Unix-like systems there is no way to hide local classes ant to prevent them from 
// being replaced by other classes with similar names in other libraries. The use of 
// unusual names may help preventing name clashes.

#ifndef _WIN32
#define Spectrum				_Local_RWDBN8_Spectrum_
#define RailwaySource			_Local_RWDBN8_RailwaySource_
#define RailwaySourcePosition	_Local_RWDBN8_SourcePosition_
#define RailwayUnit				_Local_RWDBN8_RailwayUnit_
#define RailwayTrainUnit		_Local_RWDBN8_RailwayTrainUnit_
#define RailwayTrain			_Local_RWDBN8_RailwayTrain_
#define RailwayDatabase			_Local_RWDBN8_RailwayDatabase_
#endif

struct Spectrum : public std::vector<double>
{
	Spectrum(void) //: std::vector<double>() 
	{ 
		resize (18, 0.0) ; 
	}
	Spectrum (Spectrum const& other) : std::vector<double> (other) 
	{
		assert (size() == 18) ;
	}
} ;


//typedef double Spectrum[18] ;

// ----------------------------------------------------------------------------------
/*!
 *	\brief level 1 of the database
 *	       source records describe elementary sources and their sound power
 *	       as a function of operating conditions
 */
struct RailwaySource
{
	double		height ;		/*!< height of the equivalent source */
	double		vref ;			/*!< reference speed */
	double		slope ;			/*!< speed dependency of sound power */
	int			hdir_model ;	/*!< horizontal directivity (see \ref NMPB_RailwayEmission_HorizontalDirectivity) */
	int			vdir_model ;	/*!< vertical directivity (see \ref NMPB_RailwayEmission_VerticalDirectivity) */
	Spectrum	Lw ;			/*!< sound power spectrum at the reference speed */

	RailwaySource (void) 
	: height(), vref(), slope(), hdir_model(), vdir_model(), Lw()
	{
	}
} ;

// ----------------------------------------------------------------------------------
/*!
 *	\brief auxiliary structure for assigning and positionning an elementary 
 *	       source on a railway unit
 */
struct RailwaySourcePosition
{
	IDREF		ref ;		/*!< identifier of elementary source */
	int			nb ;		/*!< number of elementary sources */
	double		pos ;		/*!< position of first source along the unit */
	double		spacing ;	/*!< spacing of sources along the unit */

	RailwaySourcePosition(void) : ref(),nb(),pos(),spacing() {}
} ;

// ----------------------------------------------------------------------------------
/*!
 *	\brief level 2 of the database
 *	       unit records describe individual rolling stock units
 */
struct RailwayUnit
{
	std::string  name ;			/*!< identifier of the unit of rolling stock*/
	double		 length ;		/*!< lenght of the unit */
	std::string  image ;		/*!< name of the picture file */
	double		 vmax ;			/*!< maximum speed of the unit in km/h */
	double		 cref ;			/*!< reflection coefficient for screen/train interaction */
	bool		 part_of_train ;	/*!< unit is not independent rolling stock */
	std::vector<RailwaySourcePosition> source ; /*!< list of elementary sources */

	RailwayUnit (void)
		: name(), length(), image(), vmax(), cref(), part_of_train(), source() 	{ }
} ;

// ----------------------------------------------------------------------------------
/*!
 *	\brief auxiliary structure for assigning sequences of units to trains 
 */
struct RailwayTrainUnit
{
	IDREF	ref ;	/*!< identifier of the unit of rolling stock */
	int		nb ;	/*!< number of units*/

	RailwayTrainUnit(void) : ref(), nb() { }
} ;

// ----------------------------------------------------------------------------------
/*!
 *	\brief level 3 of the database
 *         train records describe complete trains made up of one or more units
 */
struct RailwayTrain
{
	std::string  name ;		/*!< identifier of the train */
	std::string  image ;	/*!< name of the picture file */
	std::vector<RailwayTrainUnit> unit ; /*!< list of units */

	RailwayTrain (void) 
		: name(), image(), unit()	{ }
} ;

// ----------------------------------------------------------------------------------
/*!
 *	\brief the database of railway rolling stock
 *
 *	the database is structures at three levels : train, units and elementary sources
 *  - trains are sequences composed of units
 *  - units are defined in terms of equivalent sources
 *  - elementary sources are described in terms of sound power and direcitivity
 *
 * a railway traffic may be composed from trains and/or units.
 */
class RailwayDatabase
{
public:

	RailwayDatabase(void)
		: unit_map(), source_map(), train_map(), trace_debug(), error_code() { }

	/*! \brief load an external file into the database
     *  \param filename name of the file
	 *  \param trace if true, trace operations to stdout
	 *  \return 0 if successful, an error code otherwise
	 */
	int LoadFile (const char* filename, bool trace = false) ;

	/*! \brief get the most recent error code
	 */
	int GetErrorCode (void) { return error_code ; }

	/*! \brief dump the contents of the database to a text file
	 *  \param filename name of the text file. If the file name is a NULL pointer
	 *  or points to an empty string, the output is written to stdout.
	 */
	int Dump (const char* filename) ;

	/*! \brief enumerates the contents of the database 
	 *  \param enumProc a user defined callback function, 
     *		   called once for each selected entry in the database
	 *  \param include_types type of database entries to be enumerated
     *  \param userdata user defined data to be passed to the callback function
	 */
	int EnumEntities (EnumRailwayEntities enumProc, 
					  unsigned int include_types,
					  void* userdata) ;

	/*! \brief enumerates the elementary sources associated with a unit or a train
	 *  \param enumProc a user defined callback function, 
     *		   called once for each elementary sources
	 *  \param id identification code for train or unit
     *  \param userdata user defined data to be passed to the callback function
	 *  \param pos position along the train (for internal use only)
	 */
	int EnumSources (EnumRailwaySources enumProc,
					 IDREF id, 
					 void* userdata,
					 double pos = 0.0) ;

	/*! \brief get the elementary source record 
	 *  \param ref the identifier of the elementary source
 	 */
	RailwaySource const* getSource (IDREF ref) ;

	/*! \brief get the rolling stock unit record
	 *  \param ref the identifier of the rolling stock unit
 	 */
	RailwayUnit   const* getUnit (IDREF ref) ;

	/*! \brief get the train record
	 *  \param ref the identifier of the train
 	 */
	RailwayTrain  const* getTrain (IDREF ref) ;

private:

	/*! \brief load a unit record from an XML file node */
	bool ParseUnit (XMLNode* child) ;

	/*! \brief load a train record from an XML file node */
	bool ParseTrain (XMLNode* child) ;

	/*! \brief load an elementary source record from an XML file node */
	bool ParseSource (XMLNode* child) ;

	/*! \brief load the unit/source decomposition from an XML file node */
	bool ParseSourcePosition (XMLNode* child, RailwaySourcePosition& pos) ;

	/*! \brief load the train/unit decomposition from an XML file node */
	bool ParseTrainUnit(XMLNode* child, RailwayTrainUnit& unit) ;

	/*! \brief structure of the table of elementary sources, indexed by identifier */
	typedef std::map<IDREF, RailwaySource>  SourceMap ;

	/*! \brief structure of table of units of rolling stock, indexed by identifier */
	typedef std::map<IDREF, RailwayUnit>    UnitMap ;

	/*! \brief structure of table of trains (i.e. composed units), indexed by identifier */
	typedef std::map<IDREF, RailwayTrain>   TrainMap ;

	/*! \brief table of units, indexed by identifier */
	UnitMap		unit_map ;

	/*! \brief table of elementary sources, indexed by identifier */
	SourceMap	source_map ;

	/*! \brief table of trains, indexed by identifier */
	TrainMap	train_map ;

	/*! \brief if true, tracing to stdout is enabled */
	bool	trace_debug ;

	/*! \brief error code produced by the last operation on the database */
	int		error_code ;
} ;

#endif
