#include "RailwayEmission.h"
#include "stdlib.h"
#include <math.h>

// frequency range for NMPB-2008
// in a future version, this might be adapted for other models

static double defaultFreq[] =
{
	100, 125, 160, 
	200, 250, 315, 
	400, 500, 630, 
	800, 1000, 1250,
	1600, 2000, 2500,
	3200, 4000, 5000
} ;

// equivalent source heights for NMPB-2008
// in a future version, this might be adapted for other models

static double defaultHeight[] =
{
	0.00, 0.50, 4.00
} ;

#define TRACE_DEBUG GetOptions(OPTION_TRACE_DEBUG)

char* my_strdup (const char* s)
{
	size_t n = strlen(s) + 1 ;
	char* c = (char*) malloc (n) ;
	memcpy (c, s, n) ;
	return c ;
}
// --------------------------------------------------------------------------------------

static inline double POW10 (double x)
{
	return (x < -99) ? 0.0 : pow(10., x/10.) ;
}

static inline double LOG10 (double x)
{
	return (x <= 0) ? -99.9 : 10 * log10(x) ;
}

template <class T> inline T MIN (T const& x, T const& y)
{
	return (x < y) ? x : y ;
}

template <class T> inline T MAX (T const& x, T const& y)
{
	return (x > y) ? x : y ;
}

extern std::string getID (const char *s) ;

// --------------------------------------------------------------------------------------

// NOTE: deleteEmission / createEmission should be called whenever we change 
// the frequency range or the equivalent source heights in the emission model

void RailwaySourceModel::createEmission(void)
{
	emission.nbFreq = 18 ;
	emission.freq = defaultFreq ;

	emission.nbSources = 3 ;
	emission.source = (RailwayEquivalentSource*) 
					  malloc (emission.nbSources * sizeof (RailwayEquivalentSource)) ;

	for (unsigned int i = 0 ; i < emission.nbSources ; i++)
	{
		emission.source[i].height = defaultHeight[i] ;
		emission.source[i].Lw_dir = (double*) malloc (emission.nbFreq * sizeof(double)) ;
	}

	emission.sin_h = 0 ;
	emission.sin_v = 0 ;
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::deleteEmission(void)
{
	for (unsigned int i = 0 ; i < emission.nbSources ; i++)
	{
		free (emission.source[i].Lw_dir) ;
	}
	free (emission.source) ;
	
	emission.nbSources = 0 ;
	emission.source = 0 ;
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::createTraffic (void)
{
	traffic.component = 0 ;
	traffic.nbComponents = 0 ;
	traffic.maxComponents = 0 ;
	traffic.railCorrection = 0 ;
	traffic.nbHours = 1 ;
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::deleteTraffic (void)
{
	if (traffic.maxComponents > 0)
	{
		ClearTraffic() ;
		free (traffic.component) ;
		traffic.maxComponents = 0 ;
		traffic.component = 0 ;
	}
}

// --------------------------------------------------------------------------------------

bool RailwaySourceModel::AddTraffic (const char* id, double nb, double speed)
{
	if (id == 0 || strlen(id) == 0 || speed <= 0 || nb <= 0) return false ;

	if (traffic.nbComponents == traffic.maxComponents)
	{
		int n = traffic.maxComponents ;
		if (n == 0) n = 4 ; else n *=  2 ;
		traffic.component = (RailwayTrafficComponent*) 
							realloc (traffic.component, n * sizeof (RailwayTrafficComponent)) ;
		traffic.maxComponents = n ;
	}

	int n = traffic.nbComponents ;
	traffic.component[n].id     = my_strdup (id) ;
	traffic.component[n].speed  = speed ;
	traffic.component[n].number = nb ;

	traffic.nbComponents = n+1 ;

	return true ;
}

// --------------------------------------------------------------------------------------

RailwayEmission const* RailwaySourceModel::GetEmission (ScreenBodyInteraction* screenBodyInteraction)
{
	try
	{
		prepareEmission() ;

		prepareScreenBodyCorrection (screenBodyInteraction) ;

		if (maxInteractions > 0 && TRACE_DEBUG)
		{
			printf ("ScreenBodyInteraction correction table\n") ;
			for (unsigned int h = 0 ; h < emission.nbSources ; h++)
			{
				printf ("source height : %.2fm \n", emission.source[h].height) ;
				printf ("Freq") ;
				for (int n = 0 ; n <= maxInteractions ; ++n) printf ("\t N=%d",n) ;
				printf ("\n") ;

				for (unsigned int i = 0 ; i < emission.nbFreq ; i++)
				{
					printf ("%.0f", emission.freq[i]) ;
					for (int n = 0 ; n <= maxInteractions ; ++n)
					{
						printf ("\t%.1f", difInteraction[h][n][i]) ;
					}
					printf ("\n") ;
				}
			}
		}
		addTraffic (traffic) ;
	}
	catch (...)
	{
		return 0 ;
	}
	finaliseEmission() ;
	return &emission ;
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::prepareEmission(void)
{
	for (unsigned int i = 0 ; i < emission.nbSources ; i++)
	{
		emission.source[i].has_Lw = false ;
		for (unsigned int j = 0 ; j < emission.nbFreq ; j++)
		{
			emission.source[i].Lw_dir[j] = 0 ;
		}
	}
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::finaliseEmission(void)
{
	for (unsigned int i = 0 ; i < emission.nbSources ; i++)
	{
		double corr_rail = (emission.source[i].height < 1.0) ? traffic.railCorrection : 0.0 ;
		for (unsigned int j = 0 ; j < emission.nbFreq ; j++)
		{
			double val = emission.source[i].Lw_dir[j] ;
			if (val > 0) emission.source[i].has_Lw = true ;
			emission.source[i].Lw_dir[j] = LOG10 (val) + corr_rail ;
		}
	}
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::addTraffic (RailwayTraffic const& traffic)
{
	for (unsigned int i = 0 ; i < traffic.nbComponents ; i++)
	{
		double speed = traffic.component[i].speed ;
		double flow = traffic.component[i].number 
			        / (1000 * traffic.component[i].speed * traffic.nbHours) ;

		IDREF id = getID (traffic.component[i].id) ;

		if (TRACE_DEBUG) printf ("add traffic \n") ;

		RailwayTrain   const* train ;
		RailwayUnit    const* unit ;
		RailwaySource  const* source ;

		if ((train = db->getTrain (id)) != 0) 
		{
			if (TRACE_DEBUG) printf ("add train [%s] \n", id.c_str()) ;
			addTrain (*train, speed, flow) ;
		}
		else if ((unit = db->getUnit (id)) != 0) 
		{
			if (TRACE_DEBUG) printf ("add unit [%s] \n", id.c_str()) ;
			addUnit (*unit, speed, flow) ;
		}
		else if ((source = db->getSource (id)) != 0) 
		{
			if (TRACE_DEBUG) printf ("add source [%s] \n", id.c_str()) ;
			addSource (*source, speed, flow) ;
		}
		else
		{
			if (TRACE_DEBUG) printf ("unknown id (%s) in addTraffic \n", id.c_str()) ;
		}
	}
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::addTrain (RailwayTrain const& train, double speed, double train_flow)
{
	for (unsigned int i = 0 ; i < train.unit.size() ; i++)
	{
		double unit_flow = train_flow * train.unit[i]. nb ;

		IDREF id = train.unit[i].ref ;

		RailwayUnit    const* unit ;
		RailwaySource  const* source ;

		if ((unit = db->getUnit (id)) != 0) 
		{
			if (TRACE_DEBUG) printf ("add unit [%s] \n", id.c_str()) ;
			addUnit (*unit, speed, unit_flow) ;
		}
		else if ((source = db->getSource (id)) != 0) 
		{
			if (TRACE_DEBUG) printf ("add source [%s] \n", id.c_str()) ;
			addSource (*source, speed, unit_flow) ;
		}
		else
		{
			if (TRACE_DEBUG) printf ("unknown id (%s) in addTrain \n", id.c_str()) ;
		}
	}
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::addUnit (RailwayUnit const& unit, double speed, double unit_flow)
{
	for (unsigned int i = 0 ; i < unit.source.size() ; i++)
	{
		IDREF id = unit.source[i].ref ;

		RailwaySource  const* source ;

		if ((source = db->getSource (id)) != 0) 
		{
			if (TRACE_DEBUG) printf ("add source [%s] \n", id.c_str()) ;

			// add sources (and their image)
			for (nbInteractions = 0 ; nbInteractions <= maxInteractions ; nbInteractions++)
			{
				double source_flow = unit_flow * unit.source[i].nb ;

				// take into account reflections on the car body
				if (nbInteractions > 0) source_flow *= pow (unit.cref, nbInteractions) ;
				
				addSource (*source, speed, source_flow) ;
			}
		}
		else
		{
			if (TRACE_DEBUG) printf ("unknown id (%s) in addUnit \n", id.c_str()) ;
		}
	}
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::addSource (RailwaySource const& source, double speed, double flow)
{
	std::vector<SourceDistribution> distrib ;
	getSourceDistribution (source, distrib) ;

	if (TRACE_DEBUG) printf ("source height = %.2fm \n", source.height) ;
	if (TRACE_DEBUG) printf ("nb eq. sources = %ld \n", distrib.size());

	if ((options & OPTION_MODIFY_HEIGHTS) == 0)
	{
		size_t nb_sources = distrib.size() ;
		if (nb_sources != 1) throw InvalidHeight() ;
		int index = distrib[0].index ;
		double height = emission.source[index].height ;
		if (fabs(source.height - height) > 0.01) throw InvalidHeight() ;
	}

	double corr_speed = pow (speed / source.vref, source.slope/10.) ;
	double corr_flow  = flow ;

	for (unsigned int i = 0 ; i < emission.nbFreq ; i++)
	{
		double Lw_source = POW10 (source.Lw[i]) ;
		
		double corr_h = 1 ;
		double corr_v = 1 ;
		if (! GetOptions (OPTION_NO_DIRECTIVITY))
		{
			corr_h = getHorizontalDirectivity (source.hdir_model, emission.sin_h, emission.freq[i]) ;
			corr_v = getVerticalDirectivity (  source.vdir_model, emission.sin_v, emission.freq[i]) ;
		}

		double corr = corr_flow * corr_speed * corr_h * corr_v ;

		// cumulate sound powers per source height
		for (unsigned int j = 0 ; j < distrib.size() ; j++)
		{
			int		index  = distrib[j].index ;
			double  weight = corr * distrib[j].weight ;
			if (i == 0 && TRACE_DEBUG) 
			{
					printf ("height = %.2fm weight = %.2f%% \n", 
					emission.source[index].height, 100. * distrib[j].weight) ;
			}

			// apply car body / barrier interaction 
			// depending on source height, number of interactions and frequency
			if (nbInteractions > 0) 
			{
				double dif_direct = difInteraction[index][0][i] ;
				double dif_image  = difInteraction[index][nbInteractions][i] ;
				if (i==10 && TRACE_DEBUG) 
				{
					printf ("N=%d H=%.2f att=%.2f at %.0fHz\n",
						nbInteractions, 
						emission.source[index].height,
						dif_image,
						emission.freq[i]) ;
				}
				weight *= POW10 (dif_image - dif_direct) ;
			}
			
			emission.source[index].Lw_dir[i] += weight * Lw_source ;
		}
	}
}

// --------------------------------------------------------------------------------------

int RailwaySourceModel::getSourceDistribution (RailwaySource const& source, 
											   std::vector<SourceDistribution>& distrib)
{
	int index_low  = -1 ;
	int index_high = -1 ;
	double delta_low = 0 ;
	double delta_high = 0 ;

	// if there is only one source height in the model
	//  assign all sound power to the equivalent source at that height

	if (emission.nbSources == 1)
	{
		distrib.resize(1) ;
		distrib[0].index  = 0 ;
		distrib[0].weight = 1.00 ;
		return 1 ;
	}

	// find equivalent source above/below the source height as defined in the database

	for (unsigned int i = 0 ; i < emission.nbSources ; ++i)
	{
		// if source height match, assign all the sound power to that height

		if (fabs(source.height - emission.source[i].height) < 0.01)
		{
			distrib.resize(1) ;
			distrib[0].index  = i ;
			distrib[0].weight = 1.00 ;
			return 1 ;
		}
		else if (source.height > emission.source[i].height)
		{
			double delta = source.height - emission.source[i].height ;
			if (index_low < 0 || delta < delta_low)
			{
				index_low = i ;
				delta_low = delta ;
			}
		}
		else
		{
			double delta = emission.source[i].height - source.height ;
			if (index_high < 0 || delta < delta_high)
			{
				index_high = i ;
				delta_high = delta ;
			}
		}
	}

	// if the source is below the lowest height in model, 
	// assign all power to the lowest equivalent source
	// if the source is above the highest height in model, 
	// assign all power to the highest equivalent source
	// if the source height falls in between two equivalent sources,
	// distribute the sound power over those equivalent sources 
	// proportional to the difference in height

	if (index_high < 0)
	{
		distrib.resize(1) ;
		distrib[0].index  = index_low ;
		distrib[0].weight = 1.00 ;
	}
	else if (index_low < 0)
	{
		distrib.resize(1) ;
		distrib[0].index  = index_high ;
		distrib[0].weight = 1.00 ;
	}
	else
	{
		distrib.resize(2) ;
		distrib[0].index  = index_low ;
		distrib[0].weight = delta_high / (delta_high + delta_low) ;
		distrib[1].index  = index_high ;
		distrib[1].weight = delta_low / (delta_high + delta_low) ;
	}

	return (int) distrib.size() ;
}

// --------------------------------------------------------------------------------------

const double PI = 3.1415926 ;

double RailwaySourceModel::getHorizontalDirectivity (int model, double sin_h, double freq) 
{
	if (model == HDIR_OMNI) return 1 ;

	double cos_h = sqrt (1. - sin_h * sin_h) ;
	return 4 * cos_h / PI ;
}

// --------------------------------------------------------------------------------------

double RailwaySourceModel::getVerticalDirectivity (int model, double sin_v, double freq) 
{
	if (model == VDIR_OMNI) return 1 ;

	if (model == VDIR_HEMI) return (sin_v > 0) ? 2 : 0 ;
	
	if (sin_v < 0) return 1 ;

	double sin_2v = 2. * sin_v * sqrt (1. - sin_v * sin_v) ;

	double A = 40./3. * (2./3. * sin_2v - sin_v) ;
	double B = log10 ((freq + 600.)/200.) ;

	return POW10 (A * B) ;
}

// --------------------------------------------------------------------------------------

// calculate norm of a vector in 2D
static double getNorm (double x1, double y1)
{
	return sqrt (x1*x1 + y1*y1) ;
}

// calculate distance between two points in 2D
static double getDist (double x1, double y1, double x2, double y2)
{
	return getNorm (x1-x2, y1-y2) ;
}

// calculate path difference, positive if the line of sight is blocked, negative otherwise
static double getDelta (double x1, double y1, double x2, double y2, double x3, double y3)
{
	double delta = getDist (x1,y1,x2,y2) + getDist(x2,y2,x3,y3) - getDist(x1,y1,x3,y3) ;
	double yr = y1 + (x2 - x1) * (y3 - y1) / (x3 - x1) ;
	return (yr < y2) ? delta : -delta ;
}

// attenuation due to diffraction, Deygout approximation
static double GetDiffractionDeygout (double delta, double freq)
{
	double lambda = 334 / freq ;
	double n = 2 * delta / lambda ;
	double dn = 0 ;

	if (n < -0.25)
	{
		dn = 0 ;
	}   
	else if (n < 0)
	{
		dn = -6 + 12 * sqrt(-n) ;
	}
	else if (n < 0.25)
	{
		dn = -6 - 12 * sqrt(n) ;
	}
	else if (n < 1)
	{
		dn = -8 - 8 * sqrt(n) ;
	}
	else
	{
		dn = -16 - 10 * log10 (n) ;
	}

	return dn ;
}

// attenuation due to diffraction, NMPB08 diffraction
static double GetDiffractionNMPB08 (double delta, double freq)
{
	double lambda = 334. / freq ;
	double q = 3 + 40 * delta / lambda ;
	return (q > 1) ?  -LOG10 (q) : 0.0 ;
}

// --------------------------------------------------------------------------------------

double RailwaySourceModel::getDiffraction (double delta, double freq)
{
	return GetOptions(OPTION_DEYGOUT_DIFFRACTION) ? GetDiffractionDeygout (delta, freq)
												  : GetDiffractionNMPB08 (delta, freq) ;
}

// --------------------------------------------------------------------------------------

void RailwaySourceModel::prepareScreenBodyCorrection (ScreenBodyInteraction* screenBodyInteraction) 
{
	if (screenBodyInteraction == 0 || screenBodyInteraction->maxInteractions == 0)
	{
		maxInteractions = 0 ;
		difInteraction.clear() ;
		return ;
	}

	double dist_direct ;
	double dist_image ;

	maxInteractions = screenBodyInteraction->maxInteractions ;
	difInteraction.resize (emission.nbSources) ;
	for (int i = 0 ; i <  difInteraction.size() ; ++i)
	{
		difInteraction[i].resize (maxInteractions+1) ;
		for (int j = 0 ; j <= maxInteractions ; ++j)
		{
		 // express source, barrier top and receiver in DZ coordinates, D=0 at the (image) source
			double dS = 0 ;
			double zS = emission.source[i].height  ;
			double dB = (2*j + 1) * screenBodyInteraction->dS ;
			double zB = screenBodyInteraction->hS ;
			double dR = dB + screenBodyInteraction->dR ;
			double zR = screenBodyInteraction->hR ;

		 // spherical spreading
			double corr_dist = 0 ;
			if (j == 0)
			{
				dist_direct = getDist (dS,zS,dR,zR) ;
			}
			else
			{
				dist_image = getDist (dS,zS,dR,zR) ;
				corr_dist = 2 * LOG10 (dist_direct / dist_image) ;
			}

		 // prepare output spectrum 
			difInteraction[i][j].resize (emission.nbFreq) ;
			for (unsigned int k = 0 ; k < emission.nbFreq ; ++k) difInteraction[i][j][k] = corr_dist ;

		 // diffraction
			double delta = getDelta (dS, zS, dB, zB, dR, zR) ;
			for (unsigned int k = 0 ; k < emission.nbFreq ; ++k) 
			{
				difInteraction[i][j][k] += getDiffraction (delta, emission.freq[k]) ;
			}

			if (j == 0) continue ;

		 // absorption on the inner side of the barrier
			if (screenBodyInteraction->alpha)
			{
				for (unsigned int k = 0 ; k < emission.nbFreq ; ++k) 
				{
					difInteraction[i][j][k] += j * LOG10 (1- screenBodyInteraction->alpha[k]) ;
				}
			}

		 // retro-diffraction for (multiple) reflexions on the barrier
			for (int jp = 1 ; jp <= j ; ++jp)
			{
				double dP = (2*jp-1) * screenBodyInteraction->dS ;
				double zP = zB ;
				double dQ = dR ;
				double zQ = std::max (zB * dR / dB, zR) ; 
				double delta = -getDelta (dS,zS,dP,zP,dQ,zQ) ;
				for (unsigned int k = 0 ; k < emission.nbFreq ; ++k) 
				{
					difInteraction[i][j][k] += getDiffraction (delta, emission.freq[k]) ;
				}
			}
		}
	}
}

// --------------------------------------------------------------------------------------

/*
 * /brief extern C interface for RailwaySourceModel::CreateTraffic
 *
 */
void* NMPB08_CreateRailwayTraffic (double nb_hours) 
{
	RailwaySourceModel* model = new RailwaySourceModel ;
	model->CreateTraffic(nb_hours) ;
	return (void*) model ;
}

/*
 * /brief extern C interface for RailwaySourceModel::ClearTraffic
 *
 */
int NMPB08_ClearRailwayTraffic (void* id)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return ERROR_INVALID_ID ;
	return model->ClearTraffic() ;
}

/*
 * /brief extern C interface for RailwaySourceModel::AddTraffic
 *
 */
int NMPB08_AddRailwayTraffic (void* id, const char* unit_or_train, double number, double speed)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return ERROR_INVALID_ID ;
	return model->AddTraffic(unit_or_train, number, speed) ;
}

/*
 * /brief extern C interface for RailwaySourceModel::GetEmission
 *
 */
RailwayEmission const* NMPB08_GetRailwayEmission (void* id, ScreenBodyInteraction* screenBodyInteraction)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return 0 ;
	return model->GetEmission (screenBodyInteraction) ;
}

/*
 * /brief extern C interface for RailwaySourceModel::DeleteTraffic
 *
 */
int NMPB08_DeleteRailwayTraffic (void* id)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return ERROR_INVALID_ID ;
	delete model ;
	return 0 ; 
}

/*
 * /brief extern C interface for RailwaySourceModel::SetOptions
 *
 */
unsigned int NMPB08_SetRailwayOptions (void* id, unsigned int option, bool on_off)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return 0 ;
	return model->SetOptions (option, on_off) ;
}

/*
 * /brief extern C interface for RailwaySourceModel::GetOptions
 *
 */
unsigned int NMPB08_GetRailwayOptions (void* id, unsigned int option)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return 0 ;
	return model->GetOptions (option) ;
}

/*
 * /brief extern C interface for RailwaySourceModel::SetEmissionAngles
 *
 */
int NMPB08_SetRailwayEmissionAngles (void* id, double sin_h, double sin_v)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return 0 ;
	return model->SetEmissionAngles (sin_h, sin_v) ;
}

/*
 * /brief extern C interface for RailwaySourceModel::SetRailCorrection
 *
 */
unsigned int NMPB08_SetRailCorrection (void* id, NMPB_RailCorrectionType corr)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return 0 ;
	return model->SetRailCorrection (corr) ;
}

/*
 * /brief extern C interface for RailwaySourceModel::GetRailCorrection
 *
 */
NMPB_RailCorrectionType NMPB08_GetRailCorrection (void* id)
{
	RailwaySourceModel* model = static_cast<RailwaySourceModel*>(id) ;
	if (model == 0) return 0 ;
	return model->GetRailCorrection() ;
}
