
#ifndef _NMPB08_ROAD_INCLUDED_
#define _NMPB08_ROAD_INCLUDED_

// --------------------------------------------------------------------------------------
/*!
* \file RoadEmissionNMPB08.h
* \brief Emission model for road sources to be used in the context of the
*        NMPB-2008 prediction model for environmental noise
* \author CSTB
* \version 1.0
*
* \mainpage NMPB08 Road Emission Model 
  \n\n
  The NMPB Road Emission Model is defined in \"Prvision du bruit routier, partie 1 - Calcul des
  missions sonores dues au traffic routier\" published by SETRA, June 2009)
  \n
  \n
  \section main Public functions defined in the RoadEmissionNMPB08 library
  \subsection sub1 Calculate the overall sound power level (per meter) of the equivalent line source as a function of traffic and road surface characteristics 
  - \ref NMPB08_Lwm 
  \subsection sub2 Calculate the traction an rolling components of the sound power level of the equivalent line source as a function of traffic and road surface characteristics 
  - \ref NMPB08_Lwm_rolling
  - \ref NMPB08_Lwm_traction
  \subsection sub3 Calculate the traction an rolling components of the sound power level of a single vehicle as a function of traffic conditions and road surface characteristics 
 - \ref NMPB08_Lw_rolling
 - \ref NMPB08_Lw_traction
 \subsection sub4 Declare a user-defined road surface type
 - \ref NMPB08_DefineRoadSurfaceType
*/

// --------------------------------------------------------------------------------------

#ifndef _COMPILE_NMPB
	#ifdef _WIN32 
		#define _COMPILE_NMPB extern "C" __declspec(dllexport)
	#else
		#define _COMPILE_NMPB extern "C"
	#endif 
#endif

// ==== constants ===================================

/*!
 * \brief Vehicle type
 */
enum RoadVehicleType
{
	/*!
	 * \brief Light vehicle
	 */
	VehicleType_VL = 1,
	/*!
	 * \brief Heavy goods vehicle
	 */
	VehicleType_PL = 2 
} ;

/*!
* \brief Flow type
*/
enum RoadFlowType
{
	/*!
	* \brief Constant flow
	*/
	FlowType_CONST	= 0,
	/*!
	* \brief Acceleration section 
	*/
	FlowType_ACC	= 1,
	/*!
	* \brief Deceleration section
	*/
	FlowType_DEC	= 2,
	/*!
	* \brief Startup section 
	*/
	FlowType_START	= 3,
	/*!
	* \brief Stopping section
	*/
	FlowType_STOP	= 4
} ;

/*!
* \brief Road surface type
*/
enum RoadSurfaceType
{
	/*!
	* \brief default type
	*/
	RoadSurface_Default  =  0,
	/*!
	* \brief R1 type
	*/
	RoadSurface_R1  =  1,
	/*!
	* \brief R2 type
	*/
	RoadSurface_R2  =  2,
	/*!
	* \brief R3 type
	*/
	RoadSurface_R3  =  3,
	/*!
	* \brief Draining R1 type
	*/
	RoadSurface_DR1 = 4,
	/*!
	* \brief Draining R2 type
	*/
	RoadSurface_DR2 = 5,
	/*!
	* \brief Draining R3 type
	*/
	RoadSurface_DR3 = 6,
	/*!
	* \brief User defined type
	*/
	RoadSurface_UserDefined = 7
} ;

// Road surfaces correspondences with the RoadSurfaceType enumeration
/*!
 * \brief BBUM 0/6 road surface type
 */
const int BBUM_0_6			= RoadSurface_R1;
/*!
* \brief BBDr 0/10 road surface type
*/
const int BBDR_0_10			= RoadSurface_DR1;
/*!
* \brief BBTM 0/6-type2 road surface type
*/
const int BBTM_0_6_type2	= RoadSurface_R1;
/*!
* \brief BBTM 0/6-type1 road surface type
*/
const int BBTM_0_6_type1	= RoadSurface_R1;
/*!
* \brief BBTM 0/10-type2 road surface type
*/
const int BBTM_0_10_type2	= RoadSurface_R1;
/*!
* \brief BBSG 0/10 road surface type
*/
const int BBSG_0_10			= RoadSurface_R2;
/*!
* \brief BBTM 0/10-type1 road surface type
*/
const int BBTM_0_10_type1	= RoadSurface_R2;
/*!
* \brief BBUM 0/10 road surface type
*/
const int BBUM_0_10			= RoadSurface_R2;
/*!
* \brief ECF road surface type
*/
const int ECF				= RoadSurface_R2;
/*!
* \brief BBSG 0/14 road surface type
*/
const int BBSG_0_14			= RoadSurface_R3;
/*!
* \brief BBTM 0/14 road surface type
*/
const int BBTM_0_14			= RoadSurface_R3;
/*!
* \brief ES 6/10 road surface type
*/
const int ES_6_10			= RoadSurface_R3;
/*!
* \brief BC road surface type
*/
const int BC				= RoadSurface_R3;
/*!
* \brief ES 10/14 road surface type
*/
const int ES_10_14			= RoadSurface_R3;


/*!
* \brief Spectrum type
*/
enum RoadSpectrumType
{
	/*!
	* \brief The global value in dB(A) : only one value
	*/
	Spectrum_dBA      = 0,
	/*!
	* \brief Octave bands (dB(A)) : 6 frequencies from 125 to 4000
	*/
	Spectrum_oct_A    = 1,
	/*!
	* \brief Third-octave bands (dB(A)) : 18 frequencies from 100 to 5000
	*/
	Spectrum_3oct_A   = 2,
	/*!
	* \brief Octave bands  : 6 frequencies from 125 to 4000
	*/
	Spectrum_oct_lin  = 3,
	/*!
	* \brief Third-octave bands : 18 frequencies from 100 to 5000
	*/
	Spectrum_3oct_lin = 4
} ;

// ======= data structures ======================

/*!
 * \brief Road traffic component
 */
struct RoadTrafficComponent
{
	/*!
	* \brief the vehicle type
	*/
	RoadVehicleType		vehicleType ;
	/*!
	 * \brief the traffic flow : hourly vehicle flow rate, in vehicles/hour.
	 */
	double			trafficFlow ;
	/*!
	* \brief the traffic speed
	*/
	double			trafficSpeed ;
	/*!
	* \brief the traffic type
	*/
	RoadFlowType		flowType ;
} ;

/*!
* \brief Road traffic description
*/
struct RoadTraffic
{
	/*!
	* \brief the road surface type
	*/
	RoadSurfaceType		surfaceType ;
	/*!
	* \brief the surface age
	*/
	double				surfaceAge ;
	/*!
	* \brief the ramp in percent ( > 0 if rise, < 0 if down)
	*/
	double				ramp ;
	/*!
	* \brief the road traffic components number
	*/
	int					nbComponents ;
	/*!
	* \brief the road traffic components data (VL and PL)
	*/
	RoadTrafficComponent	*traffic ;
	/*!
	* \brief local buffer used for output values
	*/
	double	_Lw_values[18] ;
} ;

/*!
* \brief Calculation of the Sound power level per source line meter
*
* \param roadTraffic The road traffic data
* 
* \param spectrumType The spectrum type
* 
* \returns The Sound power level per source line meter, for each third-octave band
*
* \note this function returns a pointer to an internal memory buffer maintained within the library and shared 
*	    by calls to NMPB08_Lwm_traction, NMPB08_Lwm_traction and NMPB08_Lwm. Any call to these functions will 
*		invalidate the results obtained by previous calls. It is highly recommended that applications take a 
*		copy of the return values into dedicated memory immediately after the call to any of these functions.
*/
_COMPILE_NMPB double* NMPB08_Lwm (RoadTraffic* roadTraffic, RoadSpectrumType spectrumType) ;

/*!
* \brief Calculation of the rolling component of the Sound power level per source line meter
*
* \param roadTraffic The road traffic data
* 
* \param spectrumType The spectrum type
* 
* \returns The rolling component of the Sound power level per source line meter, for each third-octave band
*
* \note this function returns a pointer to an internal memory buffer maintained within the library and shared 
*	    by calls to NMPB08_Lwm_traction, NMPB08_Lwm_traction and NMPB08_Lwm. Any call to these functions will 
*		invalidate the results obtained by previous calls. It is highly recommended that applications take a 
*		copy of the return values into dedicated memory immediately after the call to any of these functions.
*/
_COMPILE_NMPB double* NMPB08_Lwm_rolling (RoadTraffic* roadTraffic, RoadSpectrumType spectrumType) ;

/*!
* \brief Calculation of the traction component of the Sound power level per source line meter
*
* \param roadTraffic The road traffic data
* 
* \param spectrumType The spectrum type
* 
* \returns the traction component of the Sound power level per source line meter, for each third-octave band
*
* \note this function returns a pointer to an internal memory buffer maintained within the library and shared 
*	    by calls to NMPB08_Lwm_traction, NMPB08_Lwm_traction and NMPB08_Lwm. Any call to these functions will 
*		invalidate the results obtained by previous calls. It is highly recommended that applications take a 
*		copy of the return values into dedicated memory immediately after the call to any of these functions.
*/
_COMPILE_NMPB double* NMPB08_Lwm_traction (RoadTraffic* roadTraffic, RoadSpectrumType spectrumType) ;

/*!
* \brief Calculation of the rolling component of the Sound power level of a vehicle type as a function of
*        speed and road surface characteristics
* 
* \param type The vehicle type
* 
* \param vehicleSpeed The vehicle speed
* 
* \param surfaceType The surface type
* 
* \param surfaceAge The surface age
* 
* \param specType The spectrum type
* 
* \returns the rolling component of the Sound power level of the vehicle, in third octave bands
*
* \note this function returns a pointer to an internal memory buffer maintained within the library and shared 
*	    by calls to NMPB08_LW_traction and NMPB08_Lw_traction. Any call to these functions will invalidate the
*	    results obtained by previous calls. It is highly recommended that applications take a copy of the 
*		return values into dedicated memory immediately after the call to any of these functions.
*/
_COMPILE_NMPB double* NMPB08_Lw_rolling (RoadVehicleType type, 
						   double vehicleSpeed, 
						   RoadSurfaceType surfaceType, 
						   double surfaceAge,
						   RoadSpectrumType specType) ;

/*!
* \brief Calculation of the rolling component of the Sound power level of a vehicle type as a function of
*        speed, ramp and flow type
* 
* \param type The vehicle type
* 
* \param vehicleSpeed The vehicle speed
* 
* \param flowType The flow type
* 
* \param ramp The ramp
* 
* \param specType The spectrum type
* 
* \returns the traction component of the Sound power level per source line meter, for each third-octave band
*
* \note this function returns a pointer to an internal memory buffer maintained within the library and shared 
*	    by calls to NMPB08_LW_traction and NMPB08_Lw_traction. Any call to these functions will invalidate the
*	    results obtained by previous calls. It is highly recommended that applications take a copy of the 
*		return values into dedicated memory immediately after the call to any of these functions.
*/
_COMPILE_NMPB double* NMPB08_Lw_traction (RoadVehicleType type,
						   double vehicleSpeed,
						   RoadFlowType flowType,
						   double ramp,
						   RoadSpectrumType specType) ;

/*!
* \brief Definition of a new road surface type with data needed for rolling component and for spectrum
* 
* \param idRoadSurface The road surface id
* 
* \param AVL The AVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
* 
* \param BVL The BVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
* 
* \param Vref_VL The Vref_VL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
* 
* \param APL The APL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
* 
* \param BPL The BPL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
* 
* \param Vref_PL The Vref_PL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
* 
* \param spectrum The spectral distribution for this road surface (if null, non draining default values will be used)
* 
* \returns true if all OK
*/
_COMPILE_NMPB bool NMPB08_DefineRoadSurfaceType(int idRoadSurface, 
												double AVL, double BVL, double Vref_VL, 
												double APL, double BPL, double Vref_PL,
												double const* spectrum);


#endif
