/*!
* \file RoadSurfaceDescription.cpp
* \brief RoadSurfaceDescription Class used to define a road surface
* \author CSTB
* \version 1.0
*/

#include "RoadSurfaceDescription.h"
#include <math.h>
#include <cstddef>

namespace RoadSurfaceDescriptionNMPB
{
	/*!
	* \brief Initialization of RoadSurfaceDescription class
	*/
	RoadSurfaceDescription::RoadSurfaceDescription(void)
	{
		// default value from R2 surface type
		_AVL = 20.1;
		_BVL = 55.4;
		_Vref_VL = 90;
		_APL = 20;
		_BPL = 63.4;
		_Vref_PL = 80;
	}

	/*!
	 * \brief Initialization of RoadSurfaceDescription class with given data
	 *
	 * \param AVL The AVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	 *
	 * \param BVL The BVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	 *
	 * \param Vref_VL The Vref_VL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
	 *
	 * \param APL The APL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	 *
	 * \param BPL The BPL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	 *
	 * \param Vref_PL The Vref_PL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
	 *
	 * \param spectrum The spectral distribution for this road surface
	 *
	 * \param doNormalization If true, do normalization
	 */
	RoadSurfaceDescription::RoadSurfaceDescription(double AVL, double BVL, double Vref_VL, double APL, double BPL, double Vref_PL, const double* spectrum, bool doNormalization)
	{
		_AVL = AVL;
		_BVL = BVL;
		_Vref_VL = Vref_VL;
		_APL = APL;
		_BPL = BPL;
		_Vref_PL = Vref_PL;

		//spectrum
		_spectralDistributionMap.clear();
		if(spectrum == NULL) return ;
		int frequencyNumber = 18;
		int NominalMedianFrequency[] = {100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000, 5000};
		for(int i = 0; i < frequencyNumber ; i++)
		{
			_spectralDistributionMap[NominalMedianFrequency[i]] = spectrum[i];
		}

		if(doNormalization)
		{
			SpectrumNormalization();
		}
	}

	/*!
	 * \brief Spectrum normalization to 0
	 */
	void RoadSurfaceDescription::SpectrumNormalization()
	{
		if(_spectralDistributionMap.empty()) return;

		// Sum calculation
		double spectrumSum = 0;
		for (map<int,double>::iterator it = _spectralDistributionMap.begin() ; it != _spectralDistributionMap.end(); it++ )
		{
				spectrumSum += pow(10.0, it->second / 10.0);
		}

		// Values modification
		if(spectrumSum <= 0 || spectrumSum == 1) return;
		spectrumSum = 10. * log10(spectrumSum);

		for (map<int,double>::iterator it = _spectralDistributionMap.begin() ; it != _spectralDistributionMap.end(); it++ )
		{
				it->second = it->second - spectrumSum;
		}
	}
}
